#include "gspinner.h"

/*
 * GLOBALS
 */

/* channel management */
GList *Genres;
Channel *current, *None;
/* display labels and the like */
GtkWidget *lartist; 
GtkWidget *lalbum;
GtkWidget *lsong;
GtkWidget *lchannel;
GtkWidget *lmenu;
/* trplayer */
pid_t trpid;
pthread_cond_t cond = PTHREAD_COND_INITIALIZER;

/*
 * CALLBACKS: menu_activate
 */

void menu_activate(GtkObject *o, Channel *c)
{
	if (c && c != current)
	{
		if (current->id == 0) /* 0 ==> # */
		{
			current = c;
			pthread_cond_signal(&cond);
			return;
		} 
		else /* # ==> # */
		{
			current = c;
			kill(trpid, SIGKILL);
		}
	}
}

/*
 * UTILITY FUNCTIONS 
 */

FILE *fopen_home(char *path, char *perm)
{
	char buf[256];
	snprintf(buf, 255, "%s/%s", getenv("HOME"), path);
	return fopen(buf, perm);
}

int System(char *command) 
{
	int pid, status;
	if (command == 0) return 1;
	pid = fork();
	if (pid == -1) return -1;
	if (pid == 0) 
	{
		char *argv[4];
		argv[0] = "sh";
		argv[1] = "-c";
		argv[2] = command;
		argv[3] = 0;
		execve("/bin/sh", argv, environ);
		exit(127);
	}
	trpid = pid;
	do 
	{
		if (waitpid(pid, &status, 0) == -1) 
		{
			trpid = 0;
			if (errno != EINTR) return -1;
		} 
		else 
		{
			trpid = 0;
			return status;
		}
	} 
	while(1);
}

void finish(char *s)
{
	char *t = s, *r =  NULL;
	while (t && (t = strstr(t, "&amp;")))
	{
		r = strdup(t+5);
		strcpy(t+1, r);
		free(r);
	}
	t = s;
	while (t && (t = strstr(t, "&quot;")))
	{
		r = strdup(t+6);
		*t = '"';
		strcpy(t+1, r);
		free(r);
	}
	/*t = s;
	while (t && (t = strchr(t, ':'))) *t = '-';*/
	s = realloc(s, strlen(s)+1);
}

/* 
 * PSEUDO-CLASSES: Channel, Genre
 */

Channel *Channel_new(char *name, int id)
{
	Channel *c = malloc(sizeof(Channel));
	/* internals */
	c->name = strdup(name);
	c->id = id;
	/* Gtk stuff */
	c->item = gtk_menu_item_new_with_label(c->name);
	gtk_signal_connect(GTK_OBJECT(c->item), "activate", 
			GTK_SIGNAL_FUNC(menu_activate), c);
	gtk_widget_show(c->item);
	return c;
}

Genre *Genre_new(char *name)
{
	Genre *g = malloc(sizeof(Genre));
	/* internals */
	g->name = strdup(name);
	g->channels = NULL;
	/* Gtk stuff */
	g->item = gtk_menu_item_new_with_label(g->name);
	g->menu = gtk_menu_new();
	gtk_menu_item_set_submenu(GTK_MENU_ITEM(g->item), g->menu);
	gtk_widget_show(g->item);
	gtk_widget_show(g->menu);
	return g;
}

void Genre_append(Genre *g, Channel *c)
{
	if (g && c)
	{
		g->channels = g_list_append(g->channels, c);
		gtk_menu_append(GTK_MENU(g->menu), c->item);
	}
}

/*
 * GET METHODS: get_info, get_spinner, get_channels, get_current
 */

void get_info(char **rurl, char **album, char **artist, char **song)
{
	FILE *gsf;
	int count = 4;
	char buf[512];
	*rurl = *album = *artist = *song = NULL;
	/* open file to parse info */
	if (!(gsf = fopen_home(".gspinner", "r"))) return;
	/* initialize some values */
	*rurl = malloc(128);
	*album = malloc(128);
	*artist = malloc(128);
	*song = malloc(128);
	memset(*rurl, 0, 128);
	memset(*album, 0, 128);
	memset(*artist, 0, 128);
	memset(*song, 0, 128);
	/* as long as we are still looking for info */
	while (count && fgets(buf, 511, gsf))
	{
		switch (count)
		{
			case 4:
				sscanf(buf, artist_format, *artist);
				if (**artist) 
				{
					char *c;
					c = strstr(*artist, "</span>");
					if (c) *c = '\0';
					finish(*artist);
					count--;
				}
				break;
			case 3:
				sscanf(buf, song_format, *song);
				if (**song) 
				{
					char *c;
					c = strstr(*song, "</span>");
					if (c) *c = '\0';
					finish(*song);
					count--;
				}
				break;
			case 2:
				sscanf(buf, album_format, *album);
				if (**album) 
				{
					char *c;
					c = strstr(*album, "</span>");
					if (c) *c = '\0';
					finish(*album);
					count--;
				}
				break;
			case 1:
				sscanf(buf, rurl_format, *rurl);
				if (**rurl) 
				{
					char *c;
					c = strstr(*rurl, "%3F");
					if (c) *c = '\0';
					*rurl = realloc(*rurl, strlen(*rurl)+1);
					count--;
				}
				break;
			default:
		}
	}
}

void get_spinner(int seq, int id)
{
	char *buf = malloc(512);
	snprintf(buf, 511, spinner_format, WGET, getenv("HOME"), seq, 
			versionString, time(NULL), id, versionString);
	buf = realloc(buf, strlen(buf) + 1);
	System(buf);
}

GtkWidget *get_channels()
{
	FILE *f;
	char buf[81] = "\0";
	Genre *currentGenre = NULL;
	GtkWidget *menu = gtk_menu_new();
	Genres = NULL;
	/* default channel */
	None = Channel_new("None", 0);
	gtk_menu_append(GTK_MENU(menu), None->item);
	/* open config file */
	if ((f = fopen(CONFIG_PATH, "r")))
	{
		/* read each genre, then add channels to current genre */
		while (fgets(buf, 80, f))
		{
			if (*buf == 'g')
			{
				Genre *g;
				char name[65] = "\0";
				sscanf(buf, "genre %64c\n", name);
				name[strlen(name)-1] = '\0';
				g = Genre_new(name);
				gtk_menu_append(GTK_MENU(menu), g->item);
				Genres = g_list_append(Genres, g);
				currentGenre = g;
			}
			else if (*buf == 'c')
			{
				Channel *c;
				int id;
				char name[65] = "\0";
				sscanf(buf, "channel %d\t%64c\n", &id, name);
				name[strlen(name)-1] = '\0';
				c = Channel_new(name, id);
				Genre_append(currentGenre, c);
			}
		}
		fclose(f);
	}
	gtk_widget_show_all(menu);
	return menu;
}

Channel *get_current()
{
	int id = 0;
	FILE *f;
	GList *gl, *gm = Genres;
	Channel *c;
	Genre *g;
	if ((f = fopen_home(".gspinner.channel", "r")))
	{
		fscanf(f, "%d", &id);
		fclose(f);
	}
	if (id == 0) return None;
	while (gm) /* traverse each Genre */
	{
		g = (Genre *)(gm->data);
		if (g)
		{
			gl = g->channels;
			while (gl) /* traverse each Genre's channels */
			{
				c = (Channel *)(gl->data);
				if (c && (c->id == id)) 
				{
					int index = g_list_index(Genres, g) + 1;
					gtk_option_menu_set_history(
							GTK_OPTION_MENU(lmenu),
							index);
					return c;
				}
				gl = gl->next;
			}
		}
		gm = gm->next;
	}
	return None;
}

/*
 * MAIN METHODS: spawn, run_atexit, main
 */

void spawn()
{
	int seq = 0;
	pthread_mutex_t mutex = PTHREAD_MUTEX_INITIALIZER;
	char *buf, *rurl, *album, *artist, *song;
	while (1)
	{
		if (current->id == 0) 
		{
			gtk_label_set_text(GTK_LABEL(lartist), "");
			gtk_label_set_text(GTK_LABEL(lalbum), "");
			gtk_label_set_text(GTK_LABEL(lsong), "");
			gtk_label_set_text(GTK_LABEL(lchannel), "");
			pthread_cond_wait(&cond, &mutex);
		}
		get_spinner(++seq, current->id);
		get_info(&rurl, &album, &artist, &song);
		/* set labels */
		gtk_label_set_text(GTK_LABEL(lartist), artist);
		gtk_label_set_text(GTK_LABEL(lalbum), album);
		gtk_label_set_text(GTK_LABEL(lsong), song);
		gtk_label_set_text(GTK_LABEL(lchannel), current->name);
		/* free memory */
		free(artist);
		free(album);
		free(song);
		buf = malloc(256);
		snprintf(buf, 255, trplayer_format, TRPLAYER, rurl);
		buf = realloc(buf, strlen(buf) + 1);
		System(buf);
		free(buf);
		free(rurl);
	}
	pthread_exit(0);
}

void run_atexit()
{
	FILE *f;
	if (trpid) kill(trpid, SIGKILL);
	if ((f = fopen_home(".gspinner.channel", "w")))
	{
		fprintf(f, "%d", current->id);
		fclose(f);
	}
}

int main(int argc, char *argv[])
{
	char title[24];
	pthread_t thread;
	pthread_attr_t attr;
	GtkWidget *window, *table;
	/* usage */
	if (argc > 1)
	{
		int x;
		for (x = 1; x<argc; x++)
		{
			if (!strcmp(argv[x], "--version"))
				printf("gspinner version %s (%s)\n",
						VERSION, versionString);
			if (!strcmp(argv[x], "--help"))
				printf("\nWho is John Galt?\n\n");
		}
		exit(0);
	}
	/* initialize some elements */
	gtk_init(&argc, &argv);
	atexit(run_atexit);
	setenv("REALPLAYER_HOME", REALPLAYER_DIR, 1);
	Genres = NULL;
	current = NULL;
	/* create the main window */
	window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
	snprintf(title, 23, "gSpinner-%s", VERSION);
	gtk_window_set_title(GTK_WINDOW(window), title);
	gtk_window_set_default_size(GTK_WINDOW(window), 350, 100);
	gtk_window_set_policy(GTK_WINDOW(window), 0, 0, 0);
	gtk_container_set_border_width(GTK_CONTAINER(window), 5);
	gtk_signal_connect(GTK_OBJECT(window), "delete_event", gtk_main_quit, 
			NULL);
	/* create the table to store labels and the menu */
	table = gtk_table_new(5, 2, FALSE);
	gtk_container_add(GTK_CONTAINER(window), table);
	gtk_table_set_row_spacings(GTK_TABLE(table), 0);
	gtk_table_set_col_spacings(GTK_TABLE(table), 5);
	/* create the options menu */
	lmenu = gtk_option_menu_new();
	gtk_widget_set_usize(lmenu, 150, 0);
	/* make new labels */
	lartist = gtk_label_new("");
	lalbum = gtk_label_new("");
	lsong = gtk_label_new("");
	lchannel = gtk_label_new("");
	/* pack table */
	gtk_table_attach(GTK_TABLE(table), gtk_label_new("Artist:"),
			0, 1, 0, 1, GTK_SHRINK, GTK_SHRINK, 0, 2);
	gtk_table_attach(GTK_TABLE(table), gtk_label_new("Album:"),
			0, 1, 1, 2, GTK_SHRINK, GTK_SHRINK, 0, 2);
	gtk_table_attach(GTK_TABLE(table), gtk_label_new("Song:"),
			0, 1, 2, 3, GTK_SHRINK, GTK_SHRINK, 0, 2);
	gtk_table_attach(GTK_TABLE(table), gtk_label_new("Channel:"),
			0, 1, 3, 4, GTK_SHRINK, GTK_SHRINK, 0, 2);
	gtk_table_attach(GTK_TABLE(table), lartist, 
			1, 2, 0, 1, GTK_FILL, GTK_SHRINK, 0, 2);
	gtk_table_attach(GTK_TABLE(table), lalbum, 
			1, 2, 1, 2, GTK_FILL, GTK_SHRINK, 0, 2);
	gtk_table_attach(GTK_TABLE(table), lsong, 
			1, 2, 2, 3, GTK_FILL, GTK_SHRINK, 0, 2);
	gtk_table_attach(GTK_TABLE(table), lchannel, 
			1, 2, 3, 4, GTK_FILL, GTK_SHRINK, 0, 2);
	gtk_table_attach(GTK_TABLE(table), lmenu, 
			0, 2, 4, 5, GTK_FILL, GTK_SHRINK, 0, 2);
	/* create the channel listings */
	gtk_option_menu_set_menu(GTK_OPTION_MENU(lmenu), get_channels());
	/* find current saved channel (if any) */
	current = get_current();
	/* make it all visible */
	gtk_widget_show_all(window);
	/* thread junk */
	pthread_attr_init(&attr);
	if (pthread_create(&thread, &attr, (void *)&spawn, lmenu) < 0)
	{
		perror("pthread_create");
		exit(1);
	}
	gtk_main();
	return 0;
}
